
#define _NO_CVSDB_HIJACK_
#include <assert.h>
#include "cvs.h"
#include "watch.h"
#include "edit.h"
#include "fileattr.h"
#include "getline.h"
#include "buffer.h"

#include <sys/types.h>
#include <limits.h>
#ifdef __linux__
#include <db2_185.h>
#else
#include <db.h>
// #include <db2/db_185.h>
#endif
#include <fcntl.h>
#include <stdarg.h>
#include <dirent.h>

#ifdef __linux__
#define _S_IFMT __S_IFMT
#define _S_IFDIR __S_IFDIR
#define _S_IFREG __S_IFREG
#endif

#if 1
#define DPRINTF(x) TRACE x
#define EPRINTF(x) TRACE x
#define WPRINTF(x) TRACE x
#else
#define DPRINTF(x)
#define EPRINTF(x)
#define WPRINTF(x)
#endif
#define XPRINTF(x) TRACE x

DB *cvsDB = NULL;
char cvsDBcurdir[2048];
static char *cvsDBcurfile = NULL;
int cvsDBcurfile_s = 0;
int cvsDBfilerw;
static char *cvsDBfilebuf = NULL;
static int cvsDBfilebuf_s = 0;
static char *cvsDBdirbuf = NULL;
static int cvsDBdirbuf_s = 0;
int cvsDBfilepos;
int cvsDBfilemax;
static FILE cvsDBMagicFile;
static int cvsDBMagicDir;
static char *cvsDBkey = NULL, *cvsDBdata = NULL;
static int cvsDBkey_s = 0, cvsDBdata_s = 0;

extern char cvsDBcurdir[];
extern char *server_temp_dir;

char *cvsDB_manglepath(const char *, int);
#define DBROOT "//"
#define DBROOTLEN 2

#define DBFTYPE_BASE 1
#define DBFTYPE_FILE 1
#define DBFTYPE_DIR 2


#define DCOUNTLEN 10

DBT *cvsDBgetfile(const char *path);
DBT *cvsDBgetdir(const char *);
char *cvsDBsetdir(const char *, char *, int);


void
cvsDBgetbuf(char **buf, int *size, int min)
{

	min++;
	if (min > *size) {
		*size = (min + 1024) & ~1023;
		// fprintf(stderr, "increasing buffer %p to %d\n", *buf, *size);
		*buf = realloc(*buf, *size);
		if (*buf == NULL) {
			fprintf(stderr, "realloc\n");
			exit(1);
		}
	}
}

int
cvsDBTcpy(char *dst, DBT *src)
{

	memcpy(dst, src->data, src->size);
	dst[src->size] = 0;
	return src->size;
}

int
cvsdbopen(char *name, const char *dir)
{
	char *p;

	if (cvsDB == NULL) {
		p = malloc(strlen(dir) + 1 + strlen(name) + 1);
		if (p == NULL)
			return ENOMEM;
		sprintf(p, "%s-%s", dir, name);
		cvsDB = dbopen(p, O_CREAT|O_RDWR, 0600, DB_BTREE, NULL);
		free(p);
		if (cvsDB == NULL)
			return ENOMEM;
		strcpy(cvsDBcurdir, DBROOT);
		cvsDBgetbuf(&cvsDBcurfile, &cvsDBcurfile_s, 256);
		cvsDBcurfile[0] = 0;
		cvsDBgetbuf(&cvsDBkey, &cvsDBkey_s, 256);
		cvsDBgetbuf(&cvsDBdata, &cvsDBdata_s, 1024);
		cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, 1024);
		cvsDBgetbuf(&cvsDBdirbuf, &cvsDBdirbuf_s, 1024);
	}
	return 0;
}

int
cvsdb_isopen()
{
	if (cvsDB == NULL) {
		fprintf(stderr, "cvsDB not open\n");
		exit(1);
	}
	return 0;
}

void
cvsdbclose(int always)
{
	if (always == 0 || cvsDB != NULL) {
		cvsdb_isopen();

		cvsDB->close(cvsDB);
		cvsDB = NULL;
	}
}

int
cvsdbunlink(char *name, char *dir)
{
	char *p;
	int ret;

	p = malloc(strlen(dir) + 1 + strlen(name) + 1);
	if (p == NULL)
		return ENOMEM;
	sprintf(p, "%s-%s", dir, name);
	ret = unlink(p);
	free(p);
	return ret;
}

char *
cvsDB_manglepath(const char *path, int type) {
	char *buf, *s;

	DPRINTF((TRACE_FLOW, "manglepath %s", path));
	if (strncmp(path, DBROOT, DBROOTLEN) == 0) {
		buf = malloc(strlen(path) + 1 + 1);
		strcpy(buf, path);
		goto out;
	}
	if (path[0] == 0)
		return strdup(DBROOT);
	if (path[0] == '/') {
		if (strlen(path) < strlen(server_temp_dir) ||
		    strncmp(path, server_temp_dir, strlen(server_temp_dir))) {
			DPRINTF((TRACE_ALWAYS, "cvsDB_manglepath absolute %s", path));
			exit(1);
		}
		path += strlen(server_temp_dir);
		while (path[0] == '/' && path[1] == '.')
			path += 2;
		if (path[0] == 0)
			return strdup(DBROOT);
		if (path[0] == '/') {
			path++;
			buf = malloc(strlen(DBROOT) + strlen(path) + 1 + 1);
			sprintf(buf, "%s%s", DBROOT, path);
		} else {
			DPRINTF((TRACE_ALWAYS, "cvsDB_manglepath weird %s", path));
			buf = malloc(strlen(path) + 1 + 1);
			strcpy(buf, path);
		}
		goto out;
	}
	while (path[0] == '.') {
		if (path[1] == 0) {
			buf = malloc(strlen(cvsDBcurdir) + 1 + 1);
			strcpy(buf, cvsDBcurdir);
			goto out;
		}
		if (path[1] == '/') {
			path += 2;
			continue;
		}
		if (path[1] == '.' && (path[2] == 0 || path[2] == '/')) {
			s = strrchr(cvsDBcurdir + DBROOTLEN, '/');
			if (s)
				*s = 0;
			else if (strlen(cvsDBcurdir + DBROOTLEN) > 0)
				cvsDBcurdir[DBROOTLEN] = 0;
			else
				DPRINTF((TRACE_RUN, "cvsDB_manglepath .. beyond root"));
			path += 2;
			if (path[0] == 0)
				break;
			path++;
			continue;
		}
		break;
	}

	buf = malloc(strlen(cvsDBcurdir) + 1 + strlen(path) + 1 + 1);
	sprintf(buf, "%s%s%s", cvsDBcurdir,
	    (strlen(cvsDBcurdir) > DBROOTLEN && path[0]) ? "/" : "", path);
 out:
	while (strlen(buf) > DBROOTLEN && buf[strlen(buf) - 1] == '/')
		buf[strlen(buf) - 1] = 0;
	if (type == DBFTYPE_DIR && strlen(buf) > DBROOTLEN)
		strcat(buf, "/");
	DPRINTF((TRACE_FLOW, "manglepath -> %s", buf));
	return buf;
}

inline int 
cvsDB_isdbpath(const char *path)
{
#if 0
	const char *prefix = CVSADM "/" CVSPREFIX;
#else
	const char *prefix = "CVS/,,"; /* XXX */
#endif

	if (cvsDB == NULL)
		return 0;

	if (strlen(path) > strlen(prefix) + 2 &&
	    strncmp(path, prefix, strlen(prefix)) == 0) {
		DPRINTF((TRACE_FLOW, "cvsDB_hijack CVSPREFIX open"));
		mkdir(CVSADM, 0777);
		return 0;
	}
#if 0
	if (strcmp(cvsDBcurdir, DBROOT) == 0 && strchr(path, '/') == NULL) {
		DPRINTF((TRACE_FLOW, "cvsDB_hijack level 1 file"));
		return 0;
	}
#endif

	return ((cvsDBcurdir[0] && path[0] != '/') ||
	    (strlen(path) >= strlen(server_temp_dir) &&
		strncmp(path, server_temp_dir, strlen(server_temp_dir)) == 0));
}

int
cvsDB_hijack_p(const char *path)
{

	if (cvsDB == NULL)
		return 0;

	return (cvsDB_isdbpath(path));
}

int
cvsDB_hijack_dirp(DIR *dir)
{

	if (cvsDB == NULL)
		return 0;
	return (dir == (DIR *)&cvsDBMagicDir);
}

char *
cvsDB_expandpath(const char *path)
{
	char *ret;

	if (cvsDB_hijack_p(path) && cvsDBgetfile(path) == NULL) {
		path = cvsDB_manglepath(path, DBFTYPE_BASE);
		ret = strdup(path + DBROOTLEN);
		free((void *)path);
		return ret;
	}
	return strdup(path);
}

int
cvsDB_adddirent(const char *path)
{
	DBT *d, dnull;
	char *fname;
	char *dir, *dbp, *p, *s, *t, *f;
	int flen, ret;

	ret = 1;
	dir = NULL;
	DPRINTF((TRACE_FLOW, "cvsDB_adddirent %s", path));
	ret = cvsdb_isopen();
	if (ret)
		return ret;

	p = cvsDB_manglepath(path, DBFTYPE_BASE);
	fname = strrchr(p + DBROOTLEN, '/');
	if (fname == NULL) {
		dbp = DBROOT;
		fname = p + DBROOTLEN;
	} else {
		*fname = 0;
		fname++;
		dbp = p;
	}
	flen = strlen(fname);

	d = cvsDBgetdir(dbp);
	if (d == NULL) {
		d = &dnull;
		d->size = 0;
		d->data = "";
	}
	dir = malloc(d->size + 1 + flen + 1);
	cvsDBTcpy(dir, d);
	DPRINTF((TRACE_FLOW, "cvsDB_adddirent %s was %s, adding %s", dbp, dir,
	    fname));

	f = NULL;
	for(s = dir; *s; s = t + 1) {
		t = strchr(s, '/');
		if (t == NULL)
			t = s + strlen(s);

		if (t == s) {
			while (t[1] == '/')
				t++;
			if (flen <= t - s)
				f = s;
			continue;
		}
		if ((flen == t - s) && (strncmp(s, fname, flen) == 0)) {
			ret = 0;
			goto out; /* found */
		}
		if (*t != '/') {
			s = t;
			break;
		}
	}

	if (f)
		strcpy(f, fname);
	else {
		if (s != dir) {
			*s = '/';
			s++;
		}
		strcpy(s, fname);
	}

	DPRINTF((TRACE_FLOW, "cvsDB_adddirent %s -> %s", dbp, dir));
	if (cvsDBsetdir(dbp, dir, strlen(dir)) == NULL)
		goto out;

	ret = 0;
 out:	
	if (dir)
		free(dir);
	free(p);
	return ret;
}

int
cvsDB_deldirent(const char *path)
{
	DBT *d;
	char *fname;
	char *dir, *s, *t, *f;
	int flen, ret;

	ret = 1;
	dir = NULL;
	DPRINTF((TRACE_FLOW, "cvsDB_deldirent %s", path));
	ret = cvsdb_isopen();
	if (ret)
		return ret;

	path = cvsDB_manglepath(path, DBFTYPE_BASE);
	fname = strrchr(path + DBROOTLEN, '/');
	if (fname == NULL)
		goto out;
	*fname = 0;
	fname++;
	flen = strlen(fname);

	d = cvsDBgetdir(path);
	if (d == NULL) {
		ret = 0;
		goto out;
	}
	dir = malloc(d->size + 1);
	cvsDBTcpy(dir, d);
	DPRINTF((TRACE_FLOW, "cvsDB_deldirent %s was %s", path, dir));

	f = NULL;
	for(s = dir; *s; s = t + 1) {
		t = strchr(s, '/');
		if (t == NULL)
			t = s + strlen(s);

		if (t == s) {
			while (t[1] == '/')
				t++;
			continue;
		}
		if ((flen == t - s) && (strncmp(s, fname, flen) == 0)) {
			memset(s, '/', flen);
			goto update; /* found */
		}
		if (*t != '/') {
			s = t;
			break;
		}
	}

 out:	
	if (dir)
		free(dir);
	free((void *)path);
	return ret;

 update:
	DPRINTF((TRACE_FLOW, "cvsDB_deldirent %s -> %s", path, dir));
	if (cvsDBsetdir(path, dir, strlen(dir)) == NULL)
		goto out;

	ret = 0;
	goto out;
}

DBT *
cvsDBgetpath(const char *path, int type) {
	int ret;
	DBT key;
	static DBT data;
	char *p;

	DPRINTF((TRACE_FLOW, "cvsDBgetpath %s", path));
	ret = cvsdb_isopen();
	if (ret)
		return NULL;

	p = cvsDB_manglepath(path, type);

	key.data = p;
	key.size = strlen(key.data) + 1;

	if (cvsDB->get(cvsDB, &key, &data, 0) != 0) {
		DPRINTF((TRACE_FLOW, "cvsDBgetpath %s not found", key.data));
		free(p);
		return NULL;
	}
	if (data.data == NULL)
		data.data = "";
	DPRINTF((TRACE_FLOW, "cvsDBgetpath %s -> %d", p, data.size));
	free(p);
	return &data;
}

DBT *
cvsDBgetfile(const char *path)
{

	return (cvsDBgetpath(path, DBFTYPE_FILE));
}

DBT *
cvsDBgetdir(const char *path)
{

	return (cvsDBgetpath(path, DBFTYPE_DIR));
}

char *
cvsDBsetpath(const char *path, char *val, int len, int type)
{
	int ret;
	DBT key, data;
	char *p;

	DPRINTF((TRACE_FLOW, "cvsDBsetpath %s -> %d", path, len));
	ret = cvsdb_isopen();
	if (ret)
		return NULL;

	p = cvsDB_manglepath(path, type);

	key.data = p;
	key.size = strlen(key.data) + 1;

	cvsDBgetbuf(&cvsDBdata, &cvsDBdata_s, len);
	data.data = cvsDBdata;
	memcpy(data.data, val, len);
	data.size = len;
	DPRINTF((TRACE_FLOW, "cvsDB->put %p %d %p %d", key.data, key.size,
		    data.data, data.size));
	if (cvsDB->put(cvsDB, &key, &data, 0) != 0) {
		DPRINTF((TRACE_FLOW, "cvsDBsetpath %s failed", p));
		free(p);
		return NULL;
	}
	DPRINTF((TRACE_FLOW, "cvsDBsetpath %s -> %d", p, data.size));
	free(p);
	return cvsDBdata;
}

char *
cvsDBsetfile(char *path, char *val, int len)
{

	return (cvsDBsetpath(path, val, len, DBFTYPE_FILE));
}

char *
cvsDBsetdir(const char *path, char *val, int len)
{

	return (cvsDBsetpath(path, val, len, DBFTYPE_DIR));
}

int
cvsDBdelpath(const char *path, int type)
{
	int ret;
	DBT key;
	char *p;

	DPRINTF((TRACE_FLOW, "cvsDBdelpath %s", path));
	ret = cvsdb_isopen();
	if (ret)
		return -1;

	cvsDB_deldirent(path);
	p = cvsDB_manglepath(path, type);

	key.data = p;
	key.size = strlen(key.data) + 1;

	if (cvsDB->del(cvsDB, &key, 0) != 0) {
		DPRINTF((TRACE_FLOW, "cvsDBdelpath %s failed", key.data));
		free(p);
		errno = ENOENT;
		return -1;
	}
	DPRINTF((TRACE_FLOW, "cvsDBdelpath %s -> ok", key.data));
	free(p);
	return 0;
}

int
cvsDBdelfile(const char *path)
{

	return (cvsDBdelpath(path, DBFTYPE_FILE));
}

int
cvsDBdeldir(const char *path)
{

	return (cvsDBdelpath(path, DBFTYPE_DIR));
}

int
cvsDBmkdir(const char *path, mode_t mode)
{

	if (cvsDB_isdir(path))
		return 0;

	cvsDB_adddirent(path);
	DPRINTF((TRACE_FLOW, "cvsDBmkdir %s", path));
	return cvsDBsetdir(path, "", 0) != NULL ? 0 : 1;
}

int
cvsDBrmdir(const char *path)
{

	if (cvsDB_isdir(path) == 0) {
		errno = EEXIST;
		return -1;
	}

	cvsDB_deldirent(path);
	DPRINTF((TRACE_FLOW, "cvsDBrmdir %s", path));
	return cvsDBdeldir(path);
}

DIR *
cvsDBopendir(const char *name)
{
	int ret;
	DBT *d;
	DIR *dret;

	if (cvsDBdirbuf[0]) {
		DPRINTF((TRACE_FLOW, "opendir contents %s, try %s",
			    cvsDBdirbuf, name));
		exit (1);
	}
	dret = NULL;
	ret = cvsdb_isopen();
	if (ret) {
		errno = ENOENT;
		return NULL;
	}

	name = cvsDB_manglepath(name, DBFTYPE_BASE);
	DPRINTF((TRACE_FLOW, "cvsDBopendir %s", name));
	d = cvsDBgetdir(name);
	if (d == NULL) {
		errno = ENOENT;
		goto out;
	}

	cvsDBgetbuf(&cvsDBdirbuf, &cvsDBdirbuf_s, d->size);
	cvsDBTcpy(cvsDBdirbuf, d);

	dret = (DIR *)&cvsDBMagicDir;
	*(int *)&cvsDBMagicDir = 0;

	DPRINTF((TRACE_FLOW, "cvsDBopendir %s -> %p: %s", name, dret, 
		    cvsDBdirbuf));
 out:
	free((void *)name);
	return dret;
}

int
cvsDBclosedir(DIR *dir)
{

	*cvsDBdirbuf = 0;
}

struct dirent *
cvsDBreaddir(DIR *dir)
{
	static struct dirent dent;
	char *t;
	int pos;

	pos = *(int *)&cvsDBMagicDir;
	DPRINTF((TRACE_FLOW, "cvsDBreaddir pos %d", pos));

 again:
	if (cvsDBdirbuf[pos] == 0)
		return NULL;	/* eof */

	memset(&dent, 0, sizeof(struct dirent));
	t = strchr(&cvsDBdirbuf[pos], '/');
	if (t == NULL)
		t = &cvsDBdirbuf[pos] + strlen(&cvsDBdirbuf[pos]);
	if (t == &cvsDBdirbuf[pos]) {
		while (*t == '/')
			t++;
		pos = t - cvsDBdirbuf;
		goto again;
	}
	strncpy(dent.d_name, &cvsDBdirbuf[pos], t - &cvsDBdirbuf[pos]);

	if (*t == '/')
		t++;
	*(int *)&cvsDBMagicDir = t - cvsDBdirbuf;

	DPRINTF((TRACE_FLOW, "cvsDBreaddir -> %s %d", dent.d_name,
		    *(int *)&cvsDBMagicDir));
	return &dent;
}

int
cvsDBstat(const char *file_name, struct stat *buf)
{
	DBT *d;
	static int uid = -1;
	static int gid = -1;

	if (uid == -1 || gid == -1) {
		uid = geteuid();
		gid = getegid();
	}
	buf->st_mode &= ~_S_IFMT;
	d = cvsDBgetdir(file_name);
	if (d == NULL) {
		d = cvsDBgetfile(file_name);
		if (d == NULL) {
			errno = ENOENT;
			return -1;
		}
		buf->st_mode |= _S_IFREG;
	} else
		buf->st_mode |= _S_IFDIR;

	buf->st_uid = uid;
	buf->st_gid = gid;
	buf->st_mode |= ACCESSPERMS;

	buf->st_size = d->size;

	DPRINTF((TRACE_FLOW, "cvsDBstat %s -> found, dir %s", file_name,
		    buf->st_mode & _S_IFDIR ? "yes" : "no"));
	return 0;
}

int
cvsDB_isfile(char *path)
{

	DPRINTF((TRACE_FLOW, "cvsDB_isfile %s", path));
	return cvsDBgetfile(path) != NULL ? 1 : 0;
}

int
cvsDB_isdir(char *path)
{

	DPRINTF((TRACE_FLOW, "cvsDB_isdir %s", path));
	return cvsDBgetdir(path) != NULL ? 1 : 0;
}

int
cvsDBchdir(const char *path) {

	DPRINTF((TRACE_FLOW, "cvsDBchdir to %s from %s", path, cvsDBcurdir));
	if (cvsDB_isdbpath(path)) {
		path = cvsDB_manglepath(path, DBFTYPE_BASE);
		strcpy(cvsDBcurdir, path);
		DPRINTF((TRACE_FLOW, "cvsDBchdir -> %s", cvsDBcurdir));
		free((void *)path);
		return chdir(server_temp_dir);
	}
	cvsDBcurdir[0] = 0;
	return chdir(path);
}

int
cvsDB_rename_file(char *oldpath, char *newpath)
{
	DBT *d;
	char *val;
	int ret;

	DPRINTF((TRACE_FLOW, "cvsDB_rename_file %s -> %s", oldpath, newpath));

	d = cvsDBgetfile(oldpath);
	if (d == NULL)
		return 1;
	val = malloc(d->size + 1);
	cvsDBTcpy(val, d);
	ret = cvsDBsetfile(newpath, val, d->size) != NULL ? 0 : 1;
	free(val);
	cvsDBdelfile(oldpath);
	return ret;
}

FILE *
cvsDBfopen(const char *path, const char *mode)
{
	DBT *d;

	DPRINTF((TRACE_FLOW, "cvsDBfopen(%s) %s", mode, path));

	if (cvsDBcurfile[0]) {
		EPRINTF((TRACE_FLOW, "file open %s, try %s", cvsDBcurfile,
			    path));
		exit (1);
	}
	path = cvsDB_manglepath(path, DBFTYPE_BASE);
	cvsDBgetbuf(&cvsDBcurfile, &cvsDBcurfile_s, strlen(path));
	strcpy(cvsDBcurfile, path);
	free((void *)path);
	DPRINTF((TRACE_FLOW, "cvsDBfopen: %s", cvsDBcurfile));

	cvsDBfilerw = 1;

	switch (mode[0]) {
	case 'r':
		d = cvsDBgetfile(cvsDBcurfile);
		if (d == NULL) {
			errno = ENOENT;
			cvsDBcurfile[0] = 0;
			DPRINTF((TRACE_FLOW, "cvsDBfopen not found"));
			return NULL;
		}
		cvsDBfilerw = 0;
		cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, d->size);
		cvsDBTcpy(cvsDBfilebuf, d);
		cvsDBfilepos = 0;
		break;
	case 'a':
		d = cvsDBgetfile(cvsDBcurfile);
		if (d) {
			cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, d->size);
			cvsDBfilepos = cvsDBTcpy(cvsDBfilebuf, d);
			break;
		}
	case 'w':
		cvsDBsetfile(cvsDBcurfile, "", 0);
		cvsDBfilebuf[0] = 0;
		cvsDBfilepos = 0;
		break;
	}
	cvsDBfilemax = cvsDBfilepos;

	cvsDB_adddirent(cvsDBcurfile);
	DPRINTF((TRACE_FLOW, "cvsDBfopen pos %d", cvsDBfilepos));
	return &cvsDBMagicFile;
}

int
cvsDB_stat(const char *file_name, struct stat *buf)
{

	if (cvsDB_hijack_p(file_name))
		return cvsDBstat(file_name, buf);
	return stat(file_name, buf);
}

#if 0
int
cvsDB_open(const char *pathname, int flags)
{

	if (cvsDB_hijack_p(pathname))
		return cvsDBopen(pathname, buf);
	return stat(file_name, buf);
}
#endif

int
cvsDB_ismagicfile(FILE *f)
{

	// DPRINTF((TRACE_FLOW, "cvsDB_ismagicfile %p %p", f, &cvsDBMagicFile));
	return (f == &cvsDBMagicFile);
}

int
cvsDB_mkfspath(const char *path)
{
	struct stat sb;
	char *c, *dir, *p;
	int ret;

	WPRINTF((TRACE_RUN, "cvsDB_mkfspath path %s", path));
	ret = -1;
	p = cvsDB_manglepath(path, DBFTYPE_BASE);
	if (p == NULL)
		goto out;
	dir = p + DBROOTLEN;
	WPRINTF((TRACE_RUN, "cvsDB_mkfspath dir %s", dir));
	c = strrchr(dir, '/');
	if (c) {
		*c = 0;
		WPRINTF((TRACE_RUN, "cvsDB_mkfspath dir %s fname %s", dir, c + 1));
		if (stat(dir, &sb) != 0) {
			*c = '/';
			c = dir;
			while ((c = strchr(c, '/')) != NULL) {
				*c = 0;
				WPRINTF((TRACE_FLOW, "cvsDB_mkfspath create dir %s rest %s",
				    dir, c + 1));
				if (mkdir(dir, 0777) && errno != EEXIST) {
					WPRINTF((TRACE_FLOW,
					    "cvsDB_mkfspath mkdir failed %d", errno));
					goto out;
				}
				*c = '/';
				c++;
			}
		} else if (S_ISDIR(sb.st_mode) == 0)
			goto out;
		/* dir created/exists */
/*  		if (chdir(dir)) */
/*  			goto out; */
		WPRINTF((TRACE_RUN, "cvsDB_mkfspath dir %s created", dir));
	}
	ret = 0;
 out:
	if (p)
		free(p);
	if (ret == -1)
		errno = ENOENT;
	return ret;
}





int
xCVS_ACCESS(const char *pathname, int mode)
{

	WPRINTF((TRACE_FLOW, "xCVS_ACCESS %s %d", pathname, mode));
	if (cvsDB_hijack_p(pathname))
		return cvsDBgetfile(pathname) != NULL ? 0 : 1;
	return access(pathname, mode);
}

int
xCVS_CHDIR(const char *path)
{

	EPRINTF((TRACE_RUN, "xCVS_CHDIR %s", path));
	return cvsDBchdir(path);
}

FILE *
xCVS_XFOPEN(int how, const char *path, const char *mode, char *file, int line)
{
	FILE *ret;
	char *p;

	WPRINTF((TRACE_FLOW, "xCVS_XFOPEN at %s:%d", file, line));
	WPRINTF((TRACE_FLOW, "xCVS_XFOPEN %x %s %s", how, path, mode));
	if (cvsDB_hijack_p(path)) {
		if (how & XFOPEN_DB) {
			ret = cvsDBfopen(path, mode);
			if (ret)
				return ret;
		}
		if (how & XFOPEN_FS) {
			if (mode[0] != 'r')
				cvsDB_mkfspath(path);
			p = cvsDB_expandpath(path);
			WPRINTF((TRACE_FLOW, "xCVS_XFOPEN try path %s", p));
			ret = fopen(p,mode);
			free(p);
			if (ret)
				return ret;
		}
		return ret;
	}
	return fopen(path, mode);
}

FILE *
xCVS_FOPEN(const char *path, const char *mode, char *file, int line)
{

	WPRINTF((TRACE_FLOW, "xCVS_FOPEN at %s:%d", file, line));
	return xCVS_XFOPEN(XFOPEN_FS, path, mode, file, line);
}

FILE *
xCVS_ADMFOPEN(const char *path, const char *mode, char *file, int line)
{

	WPRINTF((TRACE_FLOW, "xCVS_FOPEN at %s:%d", file, line));
	return xCVS_XFOPEN(XFOPEN_DB, path, mode, file, line);
}

FILE *
xCVS_FDOPEN(int fildes, const char *mode)
{

	EPRINTF((TRACE_RUN, "xCVS_FDOPEN %d %s", fildes, mode));
	return fdopen(fildes, mode);
}

int
xCVS_MKDIR(const char *pathname, mode_t mode)
{

	WPRINTF((TRACE_FLOW, "xCVS_MKDIR %s %d", pathname, mode));
	if (cvsDB_hijack_p(pathname))
		return cvsDBmkdir(pathname, mode);
	return mkdir(pathname, mode);
}

int
xCVS_CREAT(const char *pathname, mode_t mode)
{

	EPRINTF((TRACE_RUN, "xCVS_CREAT %s %d", pathname, mode));
	return xCVS_OPEN(pathname, O_CREAT | O_TRUNC | O_WRONLY, mode);
	// return creat(pathname, mode);
}

int
xCVS_OPEN(const char *pathname, int flags, mode_t mode)
{
	char *p;
	int ret;

	WPRINTF((TRACE_RUN, "xCVS_OPEN %s %d %d", pathname, flags, mode));
	if (cvsDB_hijack_p(pathname)) {
		if (flags & (O_CREAT|O_RDWR|O_WRONLY|O_APPEND)) {
			ret = cvsDB_mkfspath(pathname);
			if (ret) {
				errno = ENOENT;
				return ret;
			}
		}
		p = cvsDB_expandpath(pathname);
		ret = open(p, flags, mode);
		free(p);
		return ret;
	}
	return open(pathname, flags, mode);
}

struct dirent *cvsDBreaddir(DIR *dir);
struct dirent *
xCVS_READDIR(DIR *dir)
{

	WPRINTF((TRACE_FLOW, "xCVS_READDIR %p", dir));
	if (cvsDB_hijack_dirp(dir))
		return cvsDBreaddir(dir);
	return readdir(dir);
}

int
xCVS_CLOSEDIR(DIR *dir)
{

	WPRINTF((TRACE_FLOW, "xCVS_CLOSEDIR %p", dir));
	if (cvsDB_hijack_dirp(dir))
		return cvsDBclosedir(dir);
	return closedir(dir);
}

DIR *
xCVS_OPENDIR(const char *name)
{

	WPRINTF((TRACE_FLOW, "xCVS_OPENDIR %s", name));
	if (cvsDB_hijack_p(name))
		return cvsDBopendir(name);
	return opendir(name);
}

int
xCVS_RENAME(const char *oldpath, const char *newpath)
{

	EPRINTF((TRACE_RUN, "xCVS_RENAME %s %s", oldpath, newpath));
	return rename(oldpath, newpath);
}

int
xCVS_RMDIR(const char *pathname)
{

	WPRINTF((TRACE_FLOW, "xCVS_RMDIR %s", pathname));
	if (cvsDB_hijack_p(pathname))
		return cvsDBrmdir(pathname);
	return rmdir(pathname);
}

int
xCVS_STAT(const char *file_name, struct stat *buf, char *_file, int line)
{
	char *file, *path;
	int ret;

	WPRINTF((TRACE_FLOW, "xCVS_STAT at %s:%d", _file, line));
	WPRINTF((TRACE_FLOW, "xCVS_STAT %s %p", file_name, buf));
	if (cvsDB_hijack_p(file_name)) {
		if (cvsDBstat(file_name, buf) == 0)
			return 0;
		path = cvsDB_manglepath(file_name, DBFTYPE_BASE);
		file = path + DBROOTLEN;
		ret = stat(file, buf);
		free(path);
		return ret;
	}
	return stat(file_name, buf);
}

int
xCVS_LSTAT(const char *file_name, struct stat *buf, char *_file, int line)
{
	char *file, *path;
	int ret;

	WPRINTF((TRACE_FLOW, "xCVS_LSTAT at %s:%d", _file, line));
	WPRINTF((TRACE_FLOW, "xCVS_LSTAT %s %p", file_name, buf));
	if (cvsDB_hijack_p(file_name)) {
		if (cvsDBstat(file_name, buf) == 0)
			return 0;
		path = cvsDB_manglepath(file_name, DBFTYPE_BASE);
		file = path + DBROOTLEN;
		ret = lstat(file, buf);
		free(path);
		return ret;
	}
	return lstat(file_name, buf);
}

int
xCVS_UNLINK(const char *pathname)
{
	char *p;
	int ret;

	WPRINTF((TRACE_FLOW, "xCVS_UNLINK %s", pathname));
	if (cvsDB_hijack_p(pathname)) {
		if (cvsDBdelfile(pathname) == 0)
			return 0;
		p = cvsDB_expandpath(pathname);
		ret = unlink(p);
		free(p);
		return ret;
	}
	return unlink(pathname);
}

int
xCVS_FNMATCH(const char *pattern, const char *string, int flags)
{

	EPRINTF((TRACE_RUN, "xCVS_FNMATCH %s %s %d", pattern, string, flags));
	return fnmatch(pattern, string, flags);
}


size_t
xCVS_FREAD(void *ptr, size_t size, size_t nmemb, FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fread(ptr, size, nmemb, stream);

	XPRINTF((TRACE_FLOW, "xCVS_FREAD %s, %p %d", cvsDBcurfile, ptr,
		    size * nmemb));
	cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, cvsDBfilepos +
	    size * nmemb);
	memcpy(ptr, cvsDBfilebuf + cvsDBfilepos, size * nmemb);
	cvsDBfilepos += size * nmemb;

	return nmemb;
}

size_t
xCVS_FWRITE(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fwrite(ptr, size, nmemb, stream);

	XPRINTF((TRACE_FLOW, "xCVS_FWRITE %s, %p %d", cvsDBcurfile, ptr,
		    size * nmemb));
	cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, cvsDBfilepos +
	    size * nmemb);
	memcpy(cvsDBfilebuf + cvsDBfilepos, ptr, size * nmemb);
	cvsDBfilepos += size * nmemb;
	cvsDBfilemax = cvsDBfilepos;

	return nmemb;
}

int
xCVS_FEOF(FILE *stream)
{

	if (cvsDB_ismagicfile(stream) == 0)
		return feof(stream);

	if (cvsDBfilebuf[cvsDBfilepos] == 0)
		return -1;

	return 0;
}

int
xCVS_FERROR(FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return ferror(stream);

	XPRINTF((TRACE_FLOW, "xCVS_FERROR %s", cvsDBcurfile));
	return 0;
}

int
xCVS_FCLOSE(FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fclose(stream);

	DPRINTF((TRACE_FLOW, "xCVS_FCLOSE %s, pos %d", cvsDBcurfile,
		    cvsDBfilepos));
	if (cvsDBfilerw)
		cvsDBsetfile(cvsDBcurfile, cvsDBfilebuf, cvsDBfilepos);
	cvsDBcurfile[0] = 0;
	cvsDBfilepos = 0;

	return 0;
}

char
*xCVS_FGETS(char *s, int size, FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fgets(s, size, stream);

	XPRINTF((TRACE_FLOW, "xCVS_fgets"));
				/* XXX not implemented */
	exit(1);
	return NULL;
}

int
xCVS_FPUTS(const char *s, FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fputs(s, stream);

	XPRINTF((TRACE_FLOW, "xCVS_fputs %s", s));
	return xCVS_FWRITE(s, strlen(s), 1, stream);
}

int
xCVS_FFLUSH(FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return fflush(stream);

	XPRINTF((TRACE_FLOW, "xCVS_FFLUSH %s", cvsDBcurfile));
	return 0;
}

int
xCVS_VFPRINTF(FILE *stream, const char *format, va_list va)
{
	int ret;

	if (stream != &cvsDBMagicFile)
		return vfprintf(stream, format, va);

	ret = vsnprintf(NULL, 0, format, va);
	DPRINTF((TRACE_FLOW, "xCVS_VFPRINTF(%d) %d\n", cvsDBfilepos, 
		    ret));
	cvsDBgetbuf(&cvsDBfilebuf, &cvsDBfilebuf_s, cvsDBfilepos + ret);
	ret = vsprintf(cvsDBfilebuf + cvsDBfilepos, format, va);
	DPRINTF((TRACE_FLOW, "xCVS_VFPRINTF(%d) %s", cvsDBfilepos, 
		    cvsDBfilebuf + cvsDBfilepos));
	if (ret > 0)
		cvsDBfilepos += ret;
	return ret;
}

int
xCVS_FPRINTF(FILE *stream, const char *format, ...)
{
	va_list va;
	int ret;

	va_start(va, format);
	ret = xCVS_VFPRINTF(stream, format, va);
	va_end(va);
	return ret;
}

int
xCVS_CHMOD(const char *path, mode_t mode)
{

	if (cvsDB_hijack_p(path))
		return 0;
	return chmod(path, mode);
}

ssize_t
xCVS_GETLINE(char **lineptr, size_t *n, FILE *stream)
{
	char *ep;
	int len;

	if (cvsDB_ismagicfile(stream) == 0)
		return getline(lineptr, n, stream);

	if (!lineptr || !n || !stream) {
		errno = EINVAL;
		return -1;
	}

	if (cvsDBfilebuf[cvsDBfilepos] == 0)
		return -1;

	ep = strchr(cvsDBfilebuf + cvsDBfilepos, '\n');
	if (ep == NULL)
		len = strlen(cvsDBfilebuf + cvsDBfilepos);
	else
		len = ep - (cvsDBfilebuf + cvsDBfilepos) + 1;

	if (*n < len + 1) {
		*lineptr = (char *)realloc(*lineptr, len + 1);
		*n = len + 1;
	}
	strncpy(*lineptr, cvsDBfilebuf + cvsDBfilepos, len);
	(*lineptr)[len] = 0;

	cvsDBfilepos += len;
	DPRINTF((TRACE_FLOW, "xCVS_GETLINE -> %s", *lineptr));

	return len;
}

int
xCVS_GETC(FILE *stream)
{

	if (stream != &cvsDBMagicFile)
		return getc(stream);

	XPRINTF((TRACE_FLOW, "xCVS_getc"));
				/* XXX not implemented */
	exit(1);
	return EOF;
}

int
xCVS_PUTC(int c, FILE *stream)
{
	char s[1];

	if (stream != &cvsDBMagicFile)
		return putc(c, stream);

	XPRINTF((TRACE_FLOW, "xCVS_fputc %c", c));
	s[0] = c;
	return xCVS_FWRITE(s, 1, 1, stream);
}

int
xCVS_FSTAT_FILENO(FILE *stream, struct stat *buf)
{
	static int uid = -1;
	static int gid = -1;

	WPRINTF((TRACE_FLOW, "xCVS_FSTAT_FILENO"));
	if (stream != &cvsDBMagicFile)
		return fstat(fileno(stream), buf);

	if (uid == -1 || gid == -1) {
		uid = geteuid();
		gid = getegid();
	}
	buf->st_mode &= ~_S_IFMT;
	buf->st_mode |= _S_IFREG;
		
	buf->st_uid = uid;
	buf->st_gid = gid;
	buf->st_mode |= ACCESSPERMS;

	buf->st_size = cvsDBfilemax;
	return 0;
}

void
TRACE(int level, const char *fmt, ...)
{
	va_list va;

	if (trace >= level) {
		fprintf(stderr, "S-> ");
		va_start(va, fmt);
		vfprintf(stderr, fmt, va);
		va_end(va);
		fprintf(stderr, "\n");
	}
}
